from rest_framework import permissions

__author__ = 'mturilin'


class IsStaffOrReadOnly(permissions.BasePermission):
    """
    Custom permission to only allow owners of an object to edit it.
    """

    def has_object_permission(self, request, view, obj):
        # Read permissions are allowed to any request,
        # so we'll always allow GET, HEAD or OPTIONS requests.
        if request.method in permissions.SAFE_METHODS:
            return True

        # Write permissions are only allowed to the staff
        return request.user.is_staff

class IsStaff(permissions.BasePermission):
    """
    Custom permission to only allow owners of an object to edit it.
    """

    def has_permission(self, request, view):
        return request.user.is_staff

    def has_object_permission(self, request, view, obj):
        # Any permissions are only allowed to the
        return request.user.is_staff


from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse
from rest_framework.status import HTTP_400_BAD_REQUEST, HTTP_404_NOT_FOUND

from easyapi.errors import BadRequestError


__author__ = 'michaelturilin'


class HttpErrorsMiddleware(object):
    """
    This class returns the following HTTP message codes:
    - Bad Request (400) for ValueError and descendants.
    - Not Found (404) for Django's ObjectDoesNotExists
    """

    def process_exception(self, request, exception):
        if isinstance(exception, (ValueError, BadRequestError)):
            return HttpResponse(exception, status=HTTP_400_BAD_REQUEST)

        if isinstance(exception, (ObjectDoesNotExist,)):
            return HttpResponse(exception, status=HTTP_404_NOT_FOUND)




import json

from easyapi.tests.factories import CompanyFactory, ProjectFactory


__author__ = 'mikhailturilin'

import pytest


@pytest.mark.django_db
def test_list_endpoint(staff_api_client):
    # create 3 companies
    for i in range(3):
        CompanyFactory()

    response = staff_api_client.get('/api/company/')
    response_data = json.loads(response.content)

    assert len(response_data) == 3


@pytest.mark.django_db
def test_instance_method_scalar_return(staff_api_client):
    # create 3 companies
    company = CompanyFactory()

    for i in range(3):
        ProjectFactory(company=company, budget=(i + 1) * 100)

    response = staff_api_client.get('/api/company/%d/total_budget/' % company.pk)
    print response.content
    response_data = json.loads(response.content)

    assert response_data == 600


@pytest.mark.django_db
@pytest.mark.parametrize("function", ["project_list", "project_qs"])
def test_instance_method_list_qs(staff_api_client, function):
    # create 3 companies
    company = CompanyFactory()

    for i in range(3):
        ProjectFactory(company=company, budget=(i + 1) * 100)

    response = staff_api_client.get('/api/company/%d/%s/' % (company.pk, function))
    response_data = json.loads(response.content)

    assert len(response_data) == 3
    for project_dict in response_data:
        check_project_dict(project_dict)


@pytest.mark.django_db
@pytest.mark.parametrize("function", ["project_list", "project_qs"])
def test_instance_method_list_qs_embed(staff_api_client, function):
    # create 3 companies
    company = CompanyFactory()

    for i in range(3):
        ProjectFactory(company=company, budget=(i + 1) * 100)

    response = staff_api_client.get('/api/company/%d/%s/' % (company.pk, function),
                                    data={'_embedded': 'company'})
    response_data = json.loads(response.content)

    assert len(response_data) == 3
    for project_dict in response_data:
        check_project_dict(project_dict)
        assert project_dict['_embedded']['company']['id'] == company.id


@pytest.mark.django_db
def test_instance_method_with_scalar_param(staff_api_client):
    # create 3 companies
    company = CompanyFactory()

    response = staff_api_client.post('/api/company/%d/multiply_by_100/' % company.pk, data={'number': 23})
    response_data = json.loads(response.content)

    assert response_data == 2300


def check_project_dict(project_dict):
    assert 'name' in project_dict
    assert 'budget' in project_dict
    assert 'start_date' in project_dict
    assert 'company_id' in project_dict
    assert 'is_open' in project_dict


@pytest.mark.django_db
def test_model_fields(staff_api_client):
    # create 3 companies
    projects = [ProjectFactory() for i in range(3)]

    response = staff_api_client.get('/api/project/%d/' % projects[0].pk)
    response_data = json.loads(response.content)

    project_dict = response_data
    check_project_dict(project_dict)


@pytest.mark.django_db
def test_string_property(staff_api_client):
    company = CompanyFactory()

    response = staff_api_client.get('/api/company/%d/' % company.pk)

    response_data = json.loads(response.content)

    assert 'title' in response_data
    assert response_data['title'] == company.title


@pytest.mark.django_db
def test_model_property(staff_api_client):
    company = CompanyFactory()
    for i in range(3):
        ProjectFactory(company=company, budget=(i + 1) * 100)

    response = staff_api_client.get('/api/company/%d/' % company.pk)

    response_data = json.loads(response.content)

    assert 'first_project' in response_data
    assert isinstance(response_data['first_project'], int)

    assert 'title' in response_data
    assert isinstance(response_data['title'], basestring)


@pytest.mark.django_db
def test_related_object_lookup(staff_api_client):
    project = ProjectFactory()

    response = staff_api_client.get('/api/project/%d/' % project.pk)

    response_data = json.loads(response.content)

    assert 'company_name' in response_data
    assert response_data['company_name'] == project.company.name





import json

import pytest
from rest_framework.status import HTTP_400_BAD_REQUEST, HTTP_200_OK

from easyapi.tests.factories import CompanyFactory, ProjectFactory, AddressFactory


__author__ = 'mikhailturilin'


@pytest.mark.django_db
def test_embedded(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/project/%d/' % projects[0].id, data={'_embedded': 'company'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    # assert '_embedded' in response_data
    assert response_data['_embedded']['company']['id'] == company.id


@pytest.mark.django_db
def test_embedded_fail(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/project/%d/' % projects[0].id, data={'_embedded': 'id'})
    assert response.status_code == HTTP_400_BAD_REQUEST


@pytest.mark.django_db
def test_embedded_in_embedded(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/project/%d/' % projects[0].id, data={'_embedded': 'company__category'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)


    company_dict = response_data['_embedded']['company']
    assert company_dict['id'] == company.id

    category_dict = company_dict['_embedded']['category']
    assert category_dict['id'] == company.category.id


@pytest.mark.django_db
def test_embedded_for_list(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/project/', data={'_embedded': 'company__category'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    project_dict = response_data[0]

    company_dict = project_dict['_embedded']['company']
    assert company_dict['id'] == company.id

    category_dict = company_dict['_embedded']['category']
    assert category_dict['id'] == company.category.id


@pytest.mark.django_db
def test_embedded_model_level(staff_api_client):

    project = ProjectFactory()

    response = staff_api_client.get('/auto-list/test_project/manager/%d/' % project.manager.id, data={'_embedded': 'projects'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)


    assert len(response_data['_embedded']['projects']) == 1



@pytest.mark.django_db
def test_embedded_related(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects']) == 3



@pytest.mark.django_db
def test_embedded_function(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects_embedded'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects_embedded']) == 3




@pytest.mark.django_db
def test_embedded_function_datatype(staff_api_client):
    company = CompanyFactory()

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'my_company_type'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert response_data['_embedded']['my_company_type'] == company.company_type.value




@pytest.mark.django_db
def test_embedded_function_custom(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects_embedded_custom_function'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects_embedded_custom_function']) == 3




@pytest.mark.django_db
def test_embedded_function_embedded(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects_embedded__company'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects_embedded']) == 3
    assert response_data['_embedded']['projects_embedded'][0]['_embedded']['company']['id'] == company.id




@pytest.mark.django_db
def test_embedded_property(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects_embedded_prop'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects_embedded_prop']) == 3




@pytest.mark.django_db
def test_embedded_property_embedded(staff_api_client):
    company = CompanyFactory()

    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects_embedded_prop__company'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects_embedded_prop']) == 3
    assert response_data['_embedded']['projects_embedded_prop'][0]['_embedded']['company']['id'] == company.id



@pytest.mark.django_db
def test_embedded_related_when_none(staff_api_client):
    company = CompanyFactory()

    # projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'projects'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert len(response_data['_embedded']['projects']) == 0



@pytest.mark.django_db
def test_embedded_one_to_one(staff_api_client):
    company = CompanyFactory(address=AddressFactory())

    # projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'address'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert response_data['_embedded']['address']['street']


@pytest.mark.django_db
def test_embedded_one_to_one_when_none(staff_api_client):
    company = CompanyFactory()

    # projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/company/%d/' % company.id, data={'_embedded': 'address'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert 'address' in response_data['_embedded']
    assert not response_data['_embedded']['address']



@pytest.mark.django_db
def test_embedded_related_one_to_one(staff_api_client):
    address = AddressFactory()
    company = CompanyFactory(address=address)

    # projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/auto-api/address/%d/' % address.id, data={'_embedded': 'company'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert response_data['_embedded']['company']['name']


@pytest.mark.django_db
def test_embedded_related_one_to_one_when_none(staff_api_client):
    address = AddressFactory()

    # projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/auto-api/address/%d/' % address.id, data={'_embedded': 'company'})
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    assert 'company' in response_data['_embedded']
    assert not response_data['_embedded']['company']





import json
from django.contrib.auth import get_user_model
from django.contrib.auth.models import User

import pytest
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED
from rest_framework.test import APIClient

from easyapi.tests.factories import CompanyFactory, ProjectFactory
from easyapi.tests.test_project.models import Company, Project


__author__ = 'mikhailturilin'


def staff_api_client():
    credentials = dict(username="ccc", password='ddd')
    try:
        User.objects.get(username=credentials['username'])
    except:
        User.objects.create_superuser(email='njnjn@njn.com', **credentials)

    client = APIClient()
    client.login(**credentials)
    return client

Company.objects.all().delete()
Project.objects.all().delete()

company = CompanyFactory()
companies =[CompanyFactory() for i in range(4)]
projects = [ProjectFactory(company=company) for i in range(3)]

response = staff_api_client().get('/api/company/')
assert response.status_code == HTTP_200_OK

response_data = json.loads(response.content)

print(response.content)









from rest_framework import pagination
from rest_framework import serializers
from easyapi.serializer import AutoModelSerializer

__author__ = 'mikhailturilin'


class EasyPaginationSerializer(pagination.PaginationSerializer):
    num_pages = serializers.Field(source='paginator.num_pages')
    number = serializers.Field(source='number')



import json

import pytest

from easyapi.tests.factories import CompanyFactory, ProjectFactory
from easyapi.tests.test_project.models import ProjectScope


__author__ = 'mikhailturilin'


@pytest.mark.django_db
def test_enum_fields_properties(staff_api_client):
    company = CompanyFactory()
    project = ProjectFactory(company=company)

    response = staff_api_client.get('/auto-list/test_project/company/%d/' % company.id)

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert 'company_type' in response_data

    assert response_data['company_type'] in [1, 2]
    assert response_data['first_project_scope'] in [scope.value for scope in ProjectScope]

@pytest.mark.django_db
def test_enum_root(staff_api_client):
    response = staff_api_client.get('/enums/')

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert 'CompanyType' in response_data
    assert response_data['CompanyType'].endswith('/enums/companytype/')

    assert 'ProjectScope' in response_data
    assert response_data['ProjectScope'].endswith('/enums/projectscope/')

@pytest.mark.django_db
def test_enum_list(staff_api_client):
    response = staff_api_client.get('/enums/companytype/')

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 2

    for first in response_data:
        assert 'type' in first
        assert 'value' in first
        assert 'full_name' in first
        assert 'name' in first

        assert first['type'] == 'CompanyType'
        assert first['name'] in ['PUBLIC', 'PRIVATE']
        assert first['full_name'].startswith('CompanyType.')
        assert first['value'] in [1,2]






from enum import Enum
from rest_framework.renderers import JSONRenderer
from rest_framework.utils.encoders import JSONEncoder
from django.db import models

from easyapi.fields import enum_value
from easyapi.serializer import serialize_instance


__author__ = 'mikhailturilin'


class ModelJSONEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, models.Model):
            return serialize_instance(o)

        elif isinstance(o, Enum):
            return enum_value(o)

        elif hasattr(o, 'to_json'):
            return o.to_json()

        else:
            return super(ModelJSONEncoder, self).default(o)


class ModelJSONRenderer(JSONRenderer):
    encoder_class = ModelJSONEncoder


from django.db.models.fields.related import ForeignKey

from rest_framework.exceptions import ParseError
from rest_framework.filters import BaseFilterBackend

from easyapi.params import smart_bool, model_param


__author__ = 'mikhail turilin'

FILTER_ALL = {
    'exact', 'iexact', 'contains', 'icontains', 'gt', 'gte', 'lt', 'lte', 'in',
    'startswith', 'istartswith', 'endswith', 'iendswith', 'range', 'year',
    'month', 'day', 'week_day', 'isnull', 'search', 'regex', 'iregex',
}

FILTER_EXACT = 'exact'
FILTER_IN = 'in'


def in_converter(base_converter):
    def inner(joined_value):
        value_list = joined_value.split(',')
        return map(base_converter, value_list)

    return inner


class Filter(object):
    def __init__(self, key, base_converter, field_path, condition):
        self.key = key
        self.condition = condition
        self.field_path = field_path
        self.base_converter = base_converter

    @property
    def full_converter(self):
        if self.condition == FILTER_IN:
            return in_converter(self.base_converter)

        return self.base_converter


def get_model_field(model, field_name):
    field = model._meta.get_field(field_name)
    return field


class QuerySetFilter(object):
    def __init__(self, queryset, request, filtering, ordering=None):
        self.ordering = ordering
        self.filtering = filtering
        self.request = request
        self.queryset = queryset

    @property
    def model_fields(self):
        return [field for field in self.get_model()._meta.fields]

    @property
    def model_fields_names(self):
        return set([field.name for field in self.model_fields])

    @property
    def all_params(self):
        return self.request.GET

    @property
    def filter_params(self):
        return {key[1:]: value for key, value in self.all_params.iteritems() if key.startswith('@')}

    def filtered_queryset(self):
        queryset = self.queryset

        for filter_key, value in self.filter_params.iteritems():
            queryset = queryset.filter(**{
                filter_key: self.get_converter(filter_key)(value)
            })

        try:
            order_by = self.all_params['order_by']
            order_fields = order_by.split(',')
            if '*' not in self.ordering and set(f.rstrip('-') for f in order_fields) <= set(self.ordering):
                raise ParseError("Invalid ordering %s" % order_by)

            queryset = queryset.order_by(*order_fields)
        except KeyError:
            pass

        return queryset


    def get_model(self):
        return self.queryset.model


    def get_main_field(self, main_field_name):
        model = self.get_model()
        field = get_model_field(model, main_field_name)
        return field


    def get_converter(self, filter_key):
        return self.get_filter(filter_key).full_converter

    def get_filter(self, filter_key):
        components = filter_key.split("__")
        main_field_name = components[0]

        # if the field is unknown - filtering is not allowed
        if main_field_name not in self.model_fields_names:
            raise ParseError("Invalid filter key %s" % filter_key)

        if len(components) == 1:
            return Filter(key=filter_key, base_converter=self.get_main_field_converter(main_field_name),
                          field_path=[main_field_name],
                          condition=[FILTER_EXACT])

        if len(components) == 2 and components[1] in self.get_conditions(main_field_name):
            return Filter(key=filter_key, base_converter=self.get_main_field_converter(main_field_name),
                          field_path=[main_field_name],
                          condition=components[1])


        # we might have foreign key chain field1__field2__field3__field4__condition
        model = self.get_model()
        field = self.get_main_field(main_field_name)

        for index, component in enumerate(components):
            is_last = index == len(components) - 1

            if is_last and index > 1 and component in FILTER_ALL:
                return Filter(key=filter_key, base_converter=self.get_field_converter(field),
                              field_path=components[:-1],
                              condition=component)

            field = get_model_field(model, component)

            if is_last:
                return Filter(key=filter_key, base_converter=self.get_field_converter(field),
                              field_path=components,
                              condition=[FILTER_EXACT])

            if not isinstance(field, ForeignKey):
                raise ParseError("Invalid filter key %s" % filter_key)

            model = field.rel.to

        raise RuntimeError("We should never be here")


    def get_main_field_converter(self, main_field_name):
        definition = self.get_filtering_def(main_field_name)
        try:
            return definition['datatype']
        except KeyError:
            return self.get_field_converter(self.get_main_field(main_field_name))


    def get_field_converter(self, field):
        internal_type = field.get_internal_type()
        if internal_type == 'BooleanField':
            return smart_bool

        if internal_type == 'IntegerField' or internal_type == 'AutoField':
            return int

        if internal_type in {'FloatField', 'DecimalField'}:
            return float

        if internal_type in {'CharField', 'TextField'}:
            return str

        if internal_type in {'ForeignKey'}:
            return model_param(field.rel.to)

        raise ParseError("Unsupported field type '%s' for model field '%s'" % (internal_type, field))

    def get_filtering_def(self, main_field_name):
        try:
            filtering_def = self.filtering[main_field_name]
        except KeyError:
            try:
                filtering_def = self.filtering['*']
            except KeyError:
                raise ParseError("Unknown field %s" % main_field_name)
        return filtering_def

    def get_conditions(self, main_field):
        try:
            return self.get_filtering_def(main_field)['filters']
        except KeyError:
            return [FILTER_EXACT]


class QuerySetFilterBackend(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):

        ordering = getattr(view, 'ordering', set())

        try:
            filtering = view.filtering
        except AttributeError:
            return queryset

        return QuerySetFilter(queryset, request, filtering, ordering).filtered_queryset()





from rest_framework.fields import Field, WritableField


__author__ = 'mikhailturilin'


class PrimaryKeyReadOnlyField(Field):
    def to_native(self, value):
        return value and value.pk


class MetaField(Field):
    def __init__(self, label=None, help_text=None):
        super(MetaField, self).__init__('*', label, help_text)

    def to_native(self, obj):
        return {
            'app': type(obj)._meta.app_label,
            'model': type(obj)._meta.model_name
        }


def enum_value(value):
    return value and value.value


class RestEnumField(WritableField):
    def __init__(self, enum=None, source=None, label=None, help_text=None, read_only=False, write_only=False,
                 required=None,
                 error_messages=None, widget=None, default=None, blank=None):
        self.enum = enum
        if not enum:
            read_only = True
        super(RestEnumField, self).__init__(source, label, help_text, read_only, write_only, required, [],
                                            error_messages, widget, default, blank)


    def to_native(self, value):
        return enum_value(value)

    def from_native(self, value):
        if not self.enum or value is None or value == '':
            return None
        for m in self.enum:
            if value == m:
                return value
            if value == m.value or str(value) == str(m.value) or str(value) == str(m):
                return m
        raise ValueError("Unknown Enum value %s" % value)





__author__ = 'mikhailturilin'

import json

import pytest

from easyapi.tests.factories import CompanyFactory


@pytest.mark.django_db
def test_list_endpoint(staff_api_client):
    for i in range(37):
        CompanyFactory()

    response = staff_api_client.get('/custom-api/company-paginator/')

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert response_data['count'] == 37
    assert response_data['num_pages'] == 8
    assert response_data['number'] == 1

    results = response_data['results']
    company = results[0]
    assert company['name']
    assert company['category_id']
    assert company['company_type']
    assert company['country']
    assert company['_meta']['app'] == 'test_project'
    assert company['_meta']['model'] == 'company'





import json

import pytest

from easyapi.tests.factories import CompanyFactory, ProjectFactory


__author__ = 'mikhailturilin'


@pytest.skip
@pytest.mark.django_db
def test_model_property(staff_api_client):
    company = CompanyFactory()
    for i in range(3):
        ProjectFactory(company=company, budget=(i + 1) * 100)

    response = staff_api_client.get('/api/company/%d/projects' % company.pk)

    assert response.status_code == 200
    response_data = json.loads(response.content)

    assert 'first_project' in response_data
    assert isinstance(response_data['first_project'], int)


from django.core.paginator import Paginator, PageNotAnInteger, EmptyPage
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.views import APIView

from easyapi.decorators import map_params
from easyapi.paginator import EasyPaginationSerializer
from easyapi.serializer import model_serializer_class
from easyapi.tests.test_project.models import Company


__author__ = 'mikhailturilin'


@api_view(['GET'])
@map_params(name=str)
def say_hello(request, name):
    return Response("Hello, {name}".format(name=name))


class WelcomeView(APIView):
    @map_params(name=str)
    def get(self, request, name, **kwargs):
        return Response("Hello, {name}".format(name=name))


class CompanyPaginator(EasyPaginationSerializer):
    class Meta:
        object_serializer_class = model_serializer_class(Company)


@api_view(['GET'])
def company_paginator(request):
    queryset = Company.objects.all()
    paginator = Paginator(queryset, 5)

    page_number = request.QUERY_PARAMS.get('page')
    try:
        page = paginator.page(page_number)
    except PageNotAnInteger:
        # If page is not an integer, deliver first page.
        page = paginator.page(1)
    except EmptyPage:
        # If page is out of range (e.g. 9999),
        # deliver last page of results.
        page = paginator.page(paginator.num_pages)

    serializer_context = {'request': request}
    serializer = CompanyPaginator(page, context=serializer_context)
    return Response(serializer.data)


from functools import wraps
from django.http.response import HttpResponseBadRequest

from rest_framework.fields import Field
from easyapi.errors import BadRequestError

from .params import extract_rest_params


__author__ = 'mikhailturilin'


def rest_method(rest_verbs=None, arg_types=None, data_type=None, many=False):
    """
    Decorator that saves the function's rest args and verbs definitions to be used later in the InstanceViewSet
    :param rest_verbs:
    :param arg_types:
    :return:
    """
    rest_verbs = rest_verbs or ['GET']
    arg_types = arg_types or {}

    def outer(function):
        function.bind_to_methods = rest_verbs
        function.arg_types = arg_types
        function.data_type = data_type
        function.many = many
        return function

    return outer

def check_unfilled_required_params(func, args, kwargs):
    unfilled = unfilled_required_params(func, args, kwargs)
    if unfilled:
        param_list_str = ", ".join(unfilled)
        raise BadRequestError("Required params (%s) are not specified" % param_list_str)


def unfilled_required_params(func, args, kwargs):
    """
    Returns the list of unfilled non-default params, which:
        - Don't have default values
        - Are not filled using args
        - Are not filled using kwargs
    """
    arguments = func.func_code.co_varnames[:func.func_code.co_argcount] # slicing to cut off * and ** args
    defaults = func.func_defaults

    num_of_defaults = len(defaults or [])

    # non default params are all params except the last "num_of_defaults"  could be empty
    non_default_args = arguments[:-num_of_defaults] if num_of_defaults else arguments

    # some the params are filled positionally with args
    unfilled_positionally_non_default_args = non_default_args[len(args):]

    # so the real unfilled params are those which are also not filled using kwargs
    unfilled_non_default_args = set(unfilled_positionally_non_default_args) - set(kwargs.keys())

    return unfilled_non_default_args



def map_params(**param_dict):
    def outer(func):
        @wraps(func)
        def inner_func(request, *args, **kwargs):
            new_kwargs = extract_rest_params(request, param_dict)
            kwargs.update(new_kwargs)

            check_unfilled_required_params(func, [request] + list(args), kwargs)

            return func(request, *args, **kwargs)

        @wraps(func)
        def inner_method(self, request, *args, **kwargs):
            new_kwargs = extract_rest_params(request, param_dict)

            kwargs.update(new_kwargs)

            check_unfilled_required_params(func, [self, request] + list(args), kwargs)

            try:
                result = func(self, request, *args, **kwargs)
            except StandardError as err:
                if isinstance(err, (ValueError, LookupError)):
                    return HttpResponseBadRequest(err)
                raise err

            return result

        if 'self' in func.func_code.co_varnames:
            return inner_method
        else:
            return inner_func

    return outer


def rest_property(property_data_type=Field, property_name=None):
    class RestProperty(Property):
        field_class = property_data_type
        name = property_name

    return RestProperty


def rest_embeddable_property(name=None, many=False, data_type=None):
    class RestProperty(Property):
        rest_embeddable_property = name
        rest_many = many
        rest_data_type = data_type

    return RestProperty


def rest_embeddable_function(name=None, many=False, data_type=None):
    def outer(func):
        func.rest_embeddable_function = name or func.__name__
        func.rest_many = many
        func.rest_data_type = data_type

        return func

    return outer


class Property(object):
    "Emulate PyProperty_Type() in Objects/descrobject.c"

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        return self.fget(obj)

    def __set__(self, obj, value):
        if self.fset is None:
            raise AttributeError("can't set attribute")
        self.fset(obj, value)

    def __delete__(self, obj):
        if self.fdel is None:
            raise AttributeError("can't delete attribute")
        self.fdel(obj)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__)

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__)

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__)


import json
import pytest
from rest_framework.status import HTTP_400_BAD_REQUEST
from easyapi.decorators import unfilled_required_params

__author__ = 'mikhailturilin'


@pytest.mark.django_db
def test_func_decorator(staff_api_client):
    response = staff_api_client.get('/custom-api/hello-func/', {'name':'Misha'})
    response_data = json.loads(response.content)

    assert 'Hello, Misha' in response_data

@pytest.mark.django_db
def test_func_decorator_no_val(staff_api_client):
    response = staff_api_client.get('/custom-api/hello-func/')
    assert response.status_code == HTTP_400_BAD_REQUEST

    assert response.content == "Required params (name) are not specified"


@pytest.mark.django_db
def test_method_decorator(staff_api_client):
    response = staff_api_client.get('/custom-api/hello-view/', {'name':'Misha'})
    assert response.status_code == 200
    response_data = json.loads(response.content)

    assert 'Hello, Misha' in response_data



@pytest.mark.django_db
def test_method_decorator_no_val(staff_api_client):
    response = staff_api_client.get('/custom-api/hello-view/')
    assert response.status_code == HTTP_400_BAD_REQUEST

    assert response.content == "Required params (name) are not specified"




def test_unfilled_non_default_params():
    def a(param1, param2, param3=1, param4=2):
        pass

    def b(param1, param2):
        pass

    assert unfilled_required_params(a, ['param1 value'], {'param4':1}) == {'param2'}
    assert unfilled_required_params(a, [], {}) == {'param1','param2'}
    assert unfilled_required_params(a, ['param1','param2'], {'param4':1}) == set()
    assert unfilled_required_params(b, ['param1'], {}) == {'param2'}




import inspect

from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import ManyToManyField
from django.utils.functional import SimpleLazyObject
from enumfields.fields import EnumFieldMixin
from rest_framework.compat import get_concrete_model
from rest_framework.exceptions import ParseError
from rest_framework.serializers import ModelSerializer, _resolve_model
from rest_framework.fields import Field
from rest_framework import fields as rest_fields

from easyapi import BottomlessDict
from easyapi.fields import MetaField, RestEnumField


__author__ = 'mikhailturilin'


def model_serializer_class(model_class):
    class _DefaultSerializer(AutoModelSerializer):
        class Meta:
            model = model_class

    return _DefaultSerializer


def model_serializer(model_class, inner_dict=None, many=False):
     return model_serializer_class(model_class)(embedded_def_dict=inner_dict, many=many)


def serialize_model(model_class, instance, inner_dict, many=False):
    if not instance:
        return None
    new_serializer = model_serializer(model_class, inner_dict, many)
    return new_serializer.to_native(instance)


def serialize_instance(instance, embedded_dict=None, many=False):
    if isinstance(instance, SimpleLazyObject):
        instance = instance._wrapped

    return serialize_model(type(instance), instance, embedded_dict, many)


def serialize_queryset(related_model, qs, inner_dict):
    new_serializer = model_serializer(related_model, inner_dict)
    return [new_serializer.to_native(inst) for inst in qs]


def convert_result(value, embedded_def_dict, data_type, many):
    if many:
        return convert_many(sequence=value, embedded_def_dict=embedded_def_dict, data_type=data_type)
    else:
        return convert_one(obj=value, embedded_def_dict=embedded_def_dict, data_type=data_type)


def convert_many(sequence, embedded_def_dict, data_type):
    return [convert_one(obj, embedded_def_dict, data_type) for obj in sequence]


def convert_one(obj, embedded_def_dict, data_type):
    if data_type:
        if issubclass(data_type, Field):
            return data_type().to_native(obj)

        if callable(data_type):
            return data_type(obj)

    if not embedded_def_dict:
        return obj

    if isinstance(obj, models.Model):
        return serialize_model(type(obj), obj, embedded_def_dict)

def update_embedded_dict(embedded_dict, embedded_params):
    if embedded_params:
        for embedded_param in embedded_params:
            components = embedded_param.split('__')
            cur_dict = embedded_dict
            for c in components:
                cur_dict = cur_dict[c]


def parse_embedded_dict(embedded_params):
    embedded_def_dict = BottomlessDict()
    update_embedded_dict(embedded_def_dict, embedded_params)
    return embedded_def_dict


def embedded_dict_from_request(request):
    embedded_param = request.QUERY_PARAMS.get('_embedded', '')
    if embedded_param:
        initial = embedded_param.split(',')
    else:
        initial = None
    return parse_embedded_dict(initial)


class EmbeddedObjectsField(Field):
    def __init__(self, model, embedded_def_dict=None):
        super(EmbeddedObjectsField, self).__init__()
        self.model = model
        self.embedded_def_dict = embedded_def_dict

        self.opts = get_concrete_model(self.model)._meta
        self.related_fields = [field for field in self.opts.fields if field.serialize if field.rel]
        self.reverse_rels = self.opts.get_all_related_objects() + self.opts.get_all_related_many_to_many_objects()

        self.related_fields_names = [field.name for field in self.related_fields]
        self.reverse_rel_names = [relation.get_accessor_name() for relation in self.reverse_rels]
        self.possible_embedded_names = set(
            self.related_fields_names +
            self.reverse_rel_names +
            self.embedded_function_names() +
            self.embedded_property_names())


    def get_embedded_def_dict(self):



        if self.embedded_def_dict:
            return self.embedded_def_dict

        try:
            embedded_def_dict = embedded_dict_from_request(self.context['request'])
        except (AttributeError, KeyError):
            # meaning no context is provided
            embedded_def_dict = BottomlessDict()


        update_embedded_dict(embedded_def_dict, getattr(self.model, 'rest_embedded', []))


        # checking that there are no unknown embedded fields
        for key in embedded_def_dict.keys():
            if key not in self.possible_embedded_names:
                raise ParseError('Unknown embedded field %s' % key)

        return embedded_def_dict


    def field_to_native(self, obj, the_field_name):
        result = {}

        embedded_def_dict = self.get_embedded_def_dict() or {}


        # for the foreign keys
        for model_field in self.related_fields:
            field_name = model_field.name

            related_model = _resolve_model(model_field.rel.to)

            if field_name in embedded_def_dict:
                result[field_name] = serialize_model(related_model,
                                                     getattr(obj, field_name),
                                                     embedded_def_dict[field_name])


        # for the reverse relations
        for relation in self.reverse_rels:
            relation_name = relation.get_accessor_name()

            if relation_name not in embedded_def_dict:
                continue

            related_model = relation.model
            to_many = relation.field.rel.multiple

            if to_many:
                result[relation_name] = serialize_queryset(related_model,
                                                           getattr(obj, relation_name).all(),
                                                           embedded_def_dict[relation_name])
            else:
                try:
                    instance = getattr(obj, relation_name)
                except ObjectDoesNotExist:
                    instance = None

                result[relation_name] = serialize_model(related_model,
                                                        instance,
                                                        embedded_def_dict[relation_name])

        # embedded functions
        for method_name, method in self.embedded_functions():
            if method_name and method_name in embedded_def_dict:
                embedded_result = method(obj)  # calling the method
                result[method_name] = convert_result(embedded_result,
                                                     embedded_def_dict[method_name],
                                                     getattr(method, 'rest_data_type', None),
                                                     getattr(method, 'rest_many', False))

        # embedded properties
        for prop_name, descriptor in self.embedded_properties():
            if prop_name and prop_name in embedded_def_dict:
                embedded_result = getattr(obj, prop_name)  # getting property value
                result[prop_name] = convert_result(embedded_result,
                                                   embedded_def_dict[prop_name],
                                                   getattr(descriptor, 'rest_data_type', None),
                                                   getattr(descriptor, 'rest_many', False))

        return result


    def embedded_function_names(self):
        return [method_name for method_name, method in self.embedded_functions()]


    def embedded_functions(self):
        return [(method.rest_embeddable_function, method) for method_name, method in
                class_functions_with_attr(self.model, 'rest_embeddable_function')]

    def embedded_property_names(self):
        return [prop_name for prop_name, descriptor in self.embedded_properties()]

    def embedded_properties(self):
        return [(getattr(descriptor, 'rest_embeddable_property') or descriptor.fget.__name__, descriptor) for
                prop_name, descriptor in class_properties_with_attr(self.model, 'rest_embeddable_property')]


def class_properties_with_attr(cls, flag_attribute):
    return inspect.getmembers(cls, predicate=lambda x: inspect.isdatadescriptor(x) and hasattr(x, flag_attribute))


def class_functions_with_attr(cls, flag_attribute):
    return inspect.getmembers(cls, predicate=lambda x: inspect.ismethod(x) and hasattr(x, flag_attribute))


class JsonField(rest_fields.Field):
    pass




class AutoModelSerializer(ModelSerializer):
    def __init__(self, instance=None, data=None, files=None, context=None, partial=False, many=None,
                 allow_add_remove=False, embedded_def_dict=None, **kwargs):
        self.embedded_def_dict = embedded_def_dict
        super(AutoModelSerializer, self).__init__(instance, data, files, context, partial, many, allow_add_remove,
                                                  **kwargs)

    def get_fields(self):
        the_fields = super(AutoModelSerializer, self).get_fields()
        model = self.opts.model

        try:
            the_fields.update(model.extra_rest_fields)
        except AttributeError:
            pass

        return the_fields

    def get_default_fields(self):
        """
        Overriding get_default_fields to do two things:
        2. Add suffix '_id' to all foreign keys
        """
        ret = super(AutoModelSerializer, self).get_default_fields()

        # Deal with forward relationships
        cls = self.opts.model
        opts = get_concrete_model(cls)._meta

        related_fields = [field for field in opts.fields if field.serialize if field.rel]

        # adding embedded fields for the foreign keys
        for model_field in related_fields:
            field_name = model_field.name

            del ret[field_name]
            to_many = isinstance(model_field, ManyToManyField)

            if not to_many:
                related_model = _resolve_model(model_field.rel.to)

                ret[field_name + '_id'] = self.get_related_field_with_source(model_field, related_model, to_many,
                                                                             source=field_name)

        # adding links field
        ret['_meta'] = MetaField()
        ret['_embedded'] = EmbeddedObjectsField(cls, embedded_def_dict=self.embedded_def_dict)



        # adding fields for properties
        for p_name, p_val in class_properties_with_attr(cls, 'field_class'):
            field_name = getattr(p_val, 'name', p_name) or p_name  # we use property name as field name by default
            ret[field_name] = p_val.field_class(source=p_name)

        return ret


    def get_related_field_with_source(self, model_field, related_model, to_many, source):
        field = self.get_related_field(model_field, related_model, to_many)
        field.source = source
        return field

    def get_field(self, model_field):
        if isinstance(model_field, EnumFieldMixin):
            return RestEnumField(model_field.enum)

        return super(AutoModelSerializer, self).get_field(model_field)




import pytest
from rest_framework.status import HTTP_401_UNAUTHORIZED, HTTP_403_FORBIDDEN, HTTP_200_OK

__author__ = 'mikhailturilin'

@pytest.mark.django_db
@pytest.mark.parametrize('url',['/auto-api/company/','/auto-api/'])
def test_no_unlogged_access(api_client, url):
    response = api_client.get(url)

    assert response.status_code in [HTTP_403_FORBIDDEN, HTTP_401_UNAUTHORIZED]




@pytest.mark.django_db
def test_no_logged_regular_user_access(user_api_client):
    response = user_api_client.get('/auto-api/company/')

    assert response.status_code in [HTTP_403_FORBIDDEN, HTTP_401_UNAUTHORIZED]


@pytest.mark.django_db
def test_staff_access(staff_api_client):
    response = staff_api_client.get('/auto-api/')

    assert response.status_code == HTTP_200_OK






import json

import pytest

from easyapi.tests.factories import CompanyFactory, ProjectFactory


__author__ = 'mikhailturilin'



@pytest.mark.django_db
def test_filtering_exact_str(staff_api_client):
    # create 3 companies with known country and 6 with random
    country = 'Prussia'
    for i in range(3):
        CompanyFactory(country=country)

    for i in range(6):
        CompanyFactory()



    response = staff_api_client.get('/api/company/', data={'@country': country})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3

@pytest.mark.django_db
def test_filtering_foreign_key(staff_api_client):
    # create 3 companies with known country and 6 with random
    company = CompanyFactory()
    for i in range(3):
        ProjectFactory(company=company)

    for i in range(6):
        ProjectFactory()


    response = staff_api_client.get('/api/project/', data={'@company': company.pk})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3


@pytest.mark.django_db
def test_filtering_subselect(staff_api_client):
    country = 'Prussia'
    company = CompanyFactory(country=country)
    for i in range(3):
        ProjectFactory(company=company)

    for i in range(6):
        ProjectFactory()


    response = staff_api_client.get('/api/project/', data={'@company__country': country})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3



@pytest.mark.django_db
def test_filtering_str_in(staff_api_client):
    # create 3 companies with known country and 6 with random
    countries = ['Prussia','Jugoslavia']
    for country in countries:
        for i in range(3):
            CompanyFactory(country=country)

    for i in range(19):
        CompanyFactory()



    response = staff_api_client.get('/api/company/', data={'@country__in': ','.join(countries)})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 6



@pytest.mark.django_db
def test_ordering(staff_api_client):
    for i in range(19):
        CompanyFactory()



    response = staff_api_client.get('/api/company/', data={'order_by': '-pk'})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    for index in range(len(response_data)-1):
        assert response_data[index]['id'] > response_data[index+1]['id']




__author__ = 'mikhailturilin'


def this_domain_url(request, location):
    if not location.startswith("/"):
        location = "/" + location
    return '%s://%s%s' % (request.is_secure() and 'https' or 'http', request.get_host(), location)




from django.db.models import get_models, get_app
from rest_framework.routers import DefaultRouter, flatten, Route, replace_methodname
from django.core.exceptions import ImproperlyConfigured
from rest_framework import views
from rest_framework.response import Response
from rest_framework.reverse import reverse

from easyapi.permissions import IsStaff
from easyapi.viewsets import InstanceViewSet


__author__ = 'mikhailturilin'

INSTANCE_METHOD_ROUTE = Route(
    url=r'^{prefix}/{lookup}/{methodname}{trailing_slash}$',
    mapping={
        '{httpmethod}': '{methodname}',
    },
    name='{basename}-{methodnamehyphen}',
    initkwargs={}
)

MANAGER_METHOD_ROUTE = Route(
    url=r'^{prefix}-manager/{methodname}{trailing_slash}$',
    mapping={
        '{httpmethod}': '{methodname}',
    },
    name='{basename}-manager-{methodnamehyphen}',
    initkwargs={}
)


class EasyApiRouter(DefaultRouter):
    permission_classes = (IsStaff,)


    def __init__(self, namespace=None, **kwargs):
        super(EasyApiRouter, self).__init__(**kwargs)
        self.namespace = namespace


    def known_actions(self):
        the_known_actions = flatten([route.mapping.values() for route in self.routes])
        return the_known_actions

    def compute_routes(self, wrapper, method_name, route):
        httpmethods = wrapper.bind_to_methods

        dynamic_routes = []
        if httpmethods:
            if method_name in self.known_actions():
                raise ImproperlyConfigured('Cannot use @rest_method decorator on '
                                           'method "%s" as it is an existing route' % method_name)
            httpmethods = [method.lower() for method in httpmethods]
            dynamic_routes.append((httpmethods, method_name))


        # Dynamic routes (@link or @action decorator)
        instance_routes = []
        for httpmethods, method_name in dynamic_routes:
            initkwargs = route.initkwargs.copy()
            # initkwargs.update(getattr(viewset, method_name).kwargs)
            instance_routes.append(Route(
                url=replace_methodname(route.url, method_name),
                mapping=dict((httpmethod, method_name) for httpmethod in httpmethods),
                name=replace_methodname(route.name, method_name),
                initkwargs=initkwargs,
            ))

        return instance_routes

    def get_routes(self, viewset):
        routes = super(EasyApiRouter, self).get_routes(viewset)

        if issubclass(viewset, InstanceViewSet):
            for method_name in viewset.instance_rest_methods():
                inst_wrapper = viewset.get_instance_method(method_name)
                instance_routes = self.compute_routes(inst_wrapper, method_name, INSTANCE_METHOD_ROUTE)

                routes.extend(instance_routes)

            for method_name in viewset.manager_rest_methods():
                inst_wrapper = viewset.get_manager_method(method_name)
                manager_routes = self.compute_routes(inst_wrapper, method_name, MANAGER_METHOD_ROUTE)

                routes.extend(manager_routes)

        return routes

    def get_method_map(self, viewset, method_map):
        is_instance_viewset = issubclass(viewset, InstanceViewSet)

        bound_methods = {}
        for method, action in method_map.items():
            if hasattr(viewset, action) \
                    or (is_instance_viewset and action in viewset.instance_rest_methods()) \
                    or (is_instance_viewset and action in viewset.manager_rest_methods()):
                bound_methods[method] = action
        return bound_methods


    def get_api_root_view(self):
        """
        I ovveriden this function to include namespaces
        """
        api_root_dict = {}
        list_name = self.routes[0].name
        for prefix, viewset, basename in self.registry:
            api_root_dict[prefix] = list_name.format(basename=basename)

        outer_self = self

        class APIRoot(views.APIView):
            _ignore_model_permissions = True
            permission_classes = outer_self.permission_classes

            def get(self, request, format=None):
                ret = {}
                for key, url_name in api_root_dict.items():
                    if outer_self.namespace:
                        url_name = "%s:%s" % (outer_self.namespace, url_name)
                    ret[key] = reverse(url_name, request=request, format=format)
                return Response(ret)

        return APIRoot.as_view()

    @property
    def urls(self):
        urls = super(EasyApiRouter, self).urls
        if self.namespace:
            return urls, self.namespace, self.namespace

        return urls


def ViewSetFactory(the_model):
    class _ModelViewSet(InstanceViewSet):
        model = the_model

    return _ModelViewSet


class AutoAppRouter(EasyApiRouter):
    def __init__(self, app_name, **kwargs):
        super(AutoAppRouter, self).__init__(**kwargs)

        app = get_app(app_name)
        for model in get_models(app):
            viewset_class = ViewSetFactory(model)
            self.register(model.__name__.lower(), viewset_class)


class AutoAppListRouter(EasyApiRouter):
    def __init__(self, *app_names, **kwargs):
        super(AutoAppListRouter, self).__init__(**kwargs)

        for app_name in app_names:
            app = get_app(app_name)
            for model in get_models(app):
                viewset_class = ViewSetFactory(model)
                prefix = "%s/%s" % (app_name, model.__name__.lower())
                self.register(prefix, viewset_class)





import inspect

from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from easyapi.encoder import ModelJSONRenderer
from easyapi.filters import QuerySetFilterBackend, FILTER_ALL
from easyapi.params import extract_rest_params
from easyapi.permissions import IsStaff
from easyapi.serializer import AutoModelSerializer, convert_result, embedded_dict_from_request


__author__ = 'mikhailturilin'


class InstanceMethodWrapper(object):
    def __init__(self, viewset, method_name):
        self.viewset = viewset
        self.method_name = method_name
        self.method = getattr(viewset.model, self.method_name)
        self.bind_to_methods = self.method.bind_to_methods
        self.arg_types = getattr(self.method, 'arg_types', {})
        self.data_type = getattr(self.method, 'data_type', None)
        self.many = getattr(self.method, 'many', False)

    def __call__(self, request, *args, **kwargs):
        instance = self.viewset.get_object()
        params = extract_rest_params(request, self.arg_types)
        result = self.method(instance, **params)
        converted_result = convert_result(result, embedded_dict_from_request(request), self.data_type, self.many)
        return Response(converted_result)


class ManagerMethodWrapper(object):
    def __init__(self, manager, method_name):
        self.method_name = method_name
        self.manager = manager
        self.method = getattr(self.manager, self.method_name)
        self.bind_to_methods = self.method.bind_to_methods
        self.arg_types = getattr(self.method, 'arg_types', {})
        self.data_type = getattr(self.method, 'data_type', None)
        self.many = getattr(self.method, 'many', False)

    def __call__(self, request, *args, **kwargs):
        params = extract_rest_params(request, self.arg_types)
        result = self.method(**params)
        converted_result = convert_result(result, embedded_dict_from_request(request), self.data_type, self.many)
        return Response(converted_result)


class InstanceViewSet(ModelViewSet):
    model_serializer_class = AutoModelSerializer

    permission_classes = (IsStaff,)

    renderer_classes = (ModelJSONRenderer, )
    filter_backend = QuerySetFilterBackend

    filtering = {
        '*': {
            'filters': FILTER_ALL
        }
    }


    @classmethod
    def instance_rest_methods(cls):
        methods_items = inspect.getmembers(cls.model, predicate=inspect.ismethod)

        for method_name, method in methods_items:
            if hasattr(method, 'bind_to_methods'):
                yield method_name

    @classmethod
    def manager_rest_methods(cls):
        methods_items = inspect.getmembers(cls.model.objects, predicate=inspect.ismethod)

        for method_name, method in methods_items:
            if hasattr(method, 'bind_to_methods'):
                yield method_name

    @classmethod
    def get_instance_method(cls, method_name):
        return getattr(cls.model, method_name)

    @classmethod
    def get_manager_method(cls, method_name):
        return getattr(cls.model.objects, method_name)

    def __init__(self, **kwargs):
        super(InstanceViewSet, self).__init__(**kwargs)

        for method_name in self.instance_rest_methods():
            setattr(self, method_name, InstanceMethodWrapper(self, method_name))

        for method_name in self.manager_rest_methods():
            setattr(self, method_name, ManagerMethodWrapper(self.model.objects, method_name))


__all__ = ['InstanceViewSet']


from django.conf.urls import patterns, url
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.reverse import reverse

__author__ = 'mikhailturilin'


def enum_view_name(enum):
    return "list-%s" % enum.__name__


def enum_list_view(enum):
    @api_view(['GET'])
    def view(request):
        return Response([enum_to_dict(enum_val) for enum_val in enum])

    return view


def enum_to_dict(enum_val):
    return {
        'name': enum_val.name,
        'value': enum_val.value,
        'full_name': str(enum_val),
        'type': type(enum_val).__name__
    }


def enum_url(enum):
    return url(r'^%s/$' % enum.__name__.lower(), enum_list_view(enum), name=enum_view_name(enum))


class EnumRouter(object):
    def __init__(self, enum_list, namespace=None):
        self.enum_list = enum_list
        self.namespace = namespace


    @property
    def urls(self):
        urls = patterns('', *self.build_urls())
        if self.namespace:
            return urls, self.namespace, self.namespace

        return urls


    def build_urls(self):
        return [enum_url(enum) for enum in self.enum_list] + [self.get_root_view_url()]

    def get_root_view_url(self):
        return url('^$', self.get_root_view())

    def get_root_view(self):
        @api_view(['GET'])
        def view(request):
            return Response({
                enum.__name__: reverse(self.enum_view_name_namespace(enum)) for enum in self.enum_list
            })

        return view

    def enum_view_name_namespace(self, enum):
        if self.namespace:
            prefix = '%s:' % self.namespace
        else:
            prefix = ''

        return prefix + enum_view_name(enum)



"""
Django settings for test_project project.

For more information on this file, see
https://docs.djangoproject.com/en/1.6/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.6/ref/settings/
"""

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
import os

BASE_DIR = os.path.dirname(os.path.dirname(__file__))


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.6/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'weicd-gs&ji7u_q6*9a9%*tge90u&%p547(tk%363sp$7o*5m6'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

TEMPLATE_DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'easyapi.tests.test_project'
)

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'easyapi.middleware.HttpErrorsMiddleware',
)

ROOT_URLCONF = 'easyapi.tests.test_project.urls'

WSGI_APPLICATION = 'test_project.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.6/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        'TEST_NAME': os.path.join(BASE_DIR, 'test_db.sqlite3'),
    }
}

# Internationalization
# https://docs.djangoproject.com/en/1.6/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.6/howto/static-files/

STATIC_URL = '/static/'

REST_FRAMEWORK = {
    # 'DEFAULT_MODEL_SERIALIZER_CLASS': 'easyapi.serializer.AutoModelSerializer',

    # 'DEFAULT_PAGINATION_SERIALIZER_CLASS': 'easyapi.paginator.EasyPaginationSerializer',
    # 'PAGINATE_BY': 10,  # Default to 10
    # 'PAGINATE_BY_PARAM': 'page_size',  # Allow client to override, using `?page_size=xxx`.
    # 'MAX_PAGINATE_BY': 100,  # Maximum limit allowed when using `?page_size=xxx`.

    'DEFAULT_AUTHENTICATION_CLASSES': (
        # 'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.SessionAuthentication',
    )
}



from datetime import date
import types

from isodate import parse_date, parse_datetime

import pytest

from easyapi.params import extract_rest_params, iso_datetime, iso_date, model_param, list_param
from easyapi.tests.factories import CompanyFactory
from easyapi.tests.test_project.models import Company


__author__ = 'mikhailturilin'


@pytest.mark.parametrize('dt_str', [
    "2014-W28",
    "2014-W28-7",
    "2014-194",
])
def test_extract_date(api_request_factory, dt_str):
    param_types = {
        'dt': iso_date
    }

    request = api_request_factory.post('/', data={'dt': dt_str}, format='json')

    params = extract_rest_params(request, param_types)

    assert 'dt' in params
    assert isinstance(params['dt'], date)
    assert params['dt'] == parse_date(dt_str)


@pytest.mark.parametrize('dt_str', [
    "2014-07-13T18:02:49+00:00",
    "2014-07-13T18:02:49Z",
])
def test_extract_datetime(api_request_factory, dt_str):
    param_types = {
        'dt': iso_datetime
    }

    request = api_request_factory.post('/', data={'dt': dt_str}, format='json')

    params = extract_rest_params(request, param_types)

    assert 'dt' in params
    assert isinstance(params['dt'], date)
    assert params['dt'] == parse_datetime(dt_str)


@pytest.mark.django_db
def test_extract_model(api_request_factory):
    company = CompanyFactory()
    param_types = {
        'company': model_param(Company)
    }

    request = api_request_factory.post('/', data={'company': company.pk}, format='json')

    params = extract_rest_params(request, param_types)

    assert 'company' in params
    assert isinstance(params['company'], Company)
    assert params['company'] == company


@pytest.mark.parametrize('list_str, result', [
    ([1, 2, 3], [1, 2, 3]), # list as list
    ("[1,2,3]", [1, 2, 3]), # list as JSON
    ("one,two,three", ['one','two','three']), # comma-separated
])
def test_extract_list(api_request_factory, list_str, result):
    param_types = {
        'my_list': list_param
    }

    request = api_request_factory.post('/', data={'my_list': list_str}, format='json')

    params = extract_rest_params(request, param_types)

    assert 'my_list' in params
    assert isinstance(params['my_list'], types.ListType)
    assert params['my_list'] == result


import json

import pytest
from rest_framework.status import HTTP_200_OK

from easyapi.tests.factories import CompanyFactory, ProjectFactory


__author__ = 'mikhailturilin'


@pytest.mark.django_db
def test_meta(staff_api_client):
    company = CompanyFactory()

    link = '/api/company/%d/' % company.id
    response = staff_api_client.get(link)
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    # assert '_embedded' in response_data

    meta = {
        'app': 'test_project',
        'model': 'company',
    }

    assert response_data['_meta'] == meta



from django.contrib.auth import get_user_model
import pytest
from rest_framework.request import Request
from rest_framework.test import APIClient, APIRequestFactory
from rest_framework.views import APIView

__author__ = 'mikhailturilin'




class RestAPIRequestFactory(APIRequestFactory):
    def request(self, **kwargs):
        request = super(APIRequestFactory, self).request(**kwargs)
        request._dont_enforce_csrf_checks = not self.enforce_csrf_checks
        return Request(request, parsers=APIView().get_parsers())


@pytest.fixture
def api_client():
    return APIClient()

@pytest.fixture
@pytest.mark.django_db
def user_api_client():
    credentials = dict(username="aaa", password='bbb')
    user = get_user_model().objects.create_user(**credentials)
    client = APIClient()
    client.login(**credentials)
    return client


@pytest.fixture
@pytest.mark.django_db
def staff_api_client():
    credentials = dict(username="ccc", password='ddd')
    user = get_user_model().objects._create_user(is_staff=True, email=None, is_superuser=False, **credentials)
    client = APIClient()
    client.login(**credentials)
    return client



@pytest.fixture
@pytest.mark.django_db
def admin_api_client():
    credentials = dict(username="eee", password='fff')
    user = get_user_model().objects.create_superuser(**credentials)
    client = APIClient()
    client.login(**credentials)
    return client


@pytest.fixture
def api_request_factory():
    return RestAPIRequestFactory()


import json

import pytest

from easyapi.tests.factories import CompanyFactory, ProjectFactory


__author__ = 'mikhailturilin'


@pytest.mark.django_db
@pytest.mark.parametrize('factory,name', [
    (CompanyFactory, 'company'),
    (ProjectFactory, 'project'),
])
def test_list_endpoint(staff_api_client, factory, name):
    for i in range(3):
        factory()

    response = staff_api_client.get('/auto-list/test_project/%s/' % name)

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3


@pytest.mark.django_db
@pytest.mark.parametrize('factory,name', [
    (CompanyFactory, 'company'),
    (ProjectFactory, 'project'),
])
def test_endpoint(staff_api_client, factory, name):
    for i in range(3):
        factory()

    response = staff_api_client.get('/auto-api/%s/' % name)

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3

@pytest.mark.django_db
@pytest.mark.parametrize('root,app_prefix', [
    ('auto-api/', ''),
    ('auto-list/', 'test_project/'),
])
def test_namespaces_for_router_root_api(staff_api_client, root, app_prefix):
    response = staff_api_client.get('/%s' % root)

    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert app_prefix + 'company' in response_data
    assert response_data[app_prefix + 'company'] == "http://testserver/%s%scompany/" % (root, app_prefix)
    assert app_prefix + 'project' in response_data
    assert response_data[app_prefix + 'project'] == "http://testserver/%s%sproject/" % (root, app_prefix)



from django.db import models
from enum import Enum
from enumfields import EnumField
from rest_framework import fields as rest_fields

from easyapi.decorators import rest_method, rest_property, rest_embeddable_function, rest_embeddable_property
from easyapi.fields import PrimaryKeyReadOnlyField, RestEnumField


__author__ = 'mikhailturilin'


class CompanyType(Enum):
    PUBLIC = 1
    PRIVATE = 2


class ProjectScope(Enum):
    Client = "Client"
    Department = "Department"
    Company = "Company"


class CompanyManager(models.Manager):
    @rest_method(arg_types={'country': str}, many=True)
    def select_by_country(self, country):
        return self.filter(country=country)


class Category(models.Model):
    name = models.TextField()


class Address(models.Model):
    street = models.TextField()


class Company(models.Model):
    name = models.TextField()
    category = models.ForeignKey(Category)
    country = models.CharField(max_length=100)
    company_type = EnumField(CompanyType)

    address = models.OneToOneField(Address, null=True, blank=True, related_name="company")

    # extra_rest_fields = {
    # # 'first_project': PrimaryKeyReadOnlyField(),
    # # 'title': Field(),
    # }

    objects = CompanyManager()

    @rest_method()
    def total_budget(self):
        return sum([float(project.budget) for project in self.projects.filter(is_open=True)])

    @rest_method(many=True) # we need many=True to support embed
    def project_list(self):
        return list(self.projects.filter(is_open=True))

    @rest_method(many=True)
    def project_qs(self):
        return self.projects.filter(is_open=True)

    @rest_method(rest_verbs=['POST'], arg_types={'number': int})
    def multiply_by_100(self, number):
        return number * 100


    @rest_embeddable_function(many=True)
    def projects_embedded(self):
        return self.projects.all()

    @rest_embeddable_function(name="projects_embedded_custom_function", many=True)
    def projects_embedded_custom(self):
        return self.projects.all()

    @rest_embeddable_property(many=True)
    def projects_embedded_prop(self):
        return self.projects.all()

    @rest_property(rest_fields.Field)
    def title(self):
        return self.name.title()

    @rest_property(PrimaryKeyReadOnlyField)
    def first_project(self):
        return self.projects.all().first()


    @rest_embeddable_function(data_type=RestEnumField)
    def my_company_type(self):
        return self.company_type

    @rest_property(RestEnumField)
    def first_project_scope(self):
        try:
            return self.first_project.scope
        except AttributeError:
            return None

class ManagerManager(models.Manager):
    @rest_method(arg_types={'id':int})
    def by_id(self, id):
        return self.get(id=id)

class Manager(models.Model):
    name = models.TextField()

    rest_embedded = ['projects']

    objects = ManagerManager()


class Project(models.Model):

    company = models.ForeignKey(Company, related_name="projects")
    name = models.TextField()
    budget = models.DecimalField(decimal_places=2, max_digits=20, default=0)
    is_open = models.BooleanField(default=True)
    start_date = models.DateField()
    manager = models.ForeignKey(Manager, related_name="projects")
    scope = EnumField(ProjectScope)

    extra_rest_fields = {
        'company_name': rest_fields.Field(source='company.name'),
    }





import json
import types

from django.db.models import Model
from django.shortcuts import get_object_or_404
from django.utils.datastructures import MergeDict
import isodate
from rest_framework.exceptions import ParseError


__author__ = 'mturilin'


def convert_param_value(param_type, param_value):
    if isinstance(param_type, Model):
        return param_type.objects.get(id=int(param_value))

    return param_type(param_value)


def extract_rest_params(request, param_types, required_params=None):
    required_params = required_params or []

    new_kwargs = {}

    request_dicts = [getattr(request, name, {}) for name in {'DATA', 'GET', 'POST'}]
    data_dict = MergeDict(*request_dicts)

    for (param_name, param_type) in param_types.iteritems():

        if param_name not in data_dict:
            if param_name in required_params:
                raise ParseError('Param "%s" is required' % param_name)
            continue

        param_value = data_dict[param_name]

        new_kwargs[param_name] = convert_param_value(param_type, param_value)

    return new_kwargs


# Converters

def smart_bool(value):
    return value.lower() in ("yes", "true", "t", "1")


iso_datetime = isodate.parse_datetime
iso_date = isodate.parse_date


def model_param(model):
    def inner(pk):
        return model.objects.get(pk=pk)

    return inner


json_param = json.loads


def list_param(list_str):
    if isinstance(list_str, types.ListType):
        return list_str

    if not isinstance(list_str, types.StringTypes):
        raise ValueError('Value of is not of the list type %s' % list_str)

    if list_str.startswith('[') and list_str.endswith(']'):
        return json.loads(list_str)

    return list_str.split(',')


def primary_key(model):
    def inner(id):
        return get_object_or_404(model, id=id)

    return inner







from django.conf.urls import patterns, include, url
from django.contrib import admin
from easyapi.enums import EnumRouter

from easyapi.router import AutoAppListRouter, AutoAppRouter
from easyapi.tests.test_project.api import router
from easyapi.tests.test_project.models import CompanyType, ProjectScope
from easyapi.tests.test_project.views import WelcomeView


admin.autodiscover()

auto_list_router = AutoAppListRouter('test_project', namespace="list")
auto_router = AutoAppRouter('test_project', namespace='normal')
enum_router = EnumRouter([CompanyType, ProjectScope])

urlpatterns = patterns(
    'easyapi.tests.test_project.views',
    # Examples:
    # url(r'^$', 'test_project.views.home', name='home'),
    # url(r'^blog/', include('blog.urls')),

    url(r'^admin/', include(admin.site.urls)),
    url(r'^api/', include(router.urls)),
    url(r'^enums/', include(enum_router.urls)),
    url(r'^auto-list/', include(auto_list_router.urls)),
    url(r'^auto-api/', include(auto_router.urls)),
    url(r'^custom-api/hello-func/', 'say_hello'),
    url(r'^custom-api/company-paginator/', 'company_paginator'),
    url(r'^custom-api/hello-view/', WelcomeView.as_view()),
)



import json
from rest_framework.status import HTTP_200_OK

from easyapi.tests.factories import CompanyFactory, ProjectFactory


__author__ = 'mikhailturilin'

import pytest


@pytest.mark.django_db
def test_manager_method(staff_api_client):
    # create 3 companies with known country and 6 with random
    country = 'Prussia'
    for i in range(3):
        CompanyFactory(country=country)

    for i in range(6):
        CompanyFactory()

    response = staff_api_client.get('/api/company-manager/select_by_country/', data={'country': country})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3


@pytest.mark.django_db
def test_manager_method_embedded(staff_api_client):
    # create 3 companies with known country and 6 with random
    country = 'Prussia'
    for i in range(3):
        c = CompanyFactory(country=country)
        for j in range(4):
            ProjectFactory(company=c)

    for i in range(6):
        CompanyFactory()

    response = staff_api_client.get('/api/company-manager/select_by_country/',
                                    data={'country': country, '_embedded': 'projects'})
    assert response.status_code == 200

    response_data = json.loads(response.content)

    assert len(response_data) == 3

    assert len(response_data[0]['_embedded']['projects']) == 4



@pytest.mark.django_db
def test_manager_method_embedded_default(staff_api_client):
    # create 3 companies with known country and 6 with random

    project = ProjectFactory()

    response = staff_api_client.get('/auto-list/test_project/manager-manager/by_id/?id=%d' % project.manager.id)
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)


    assert len(response_data['_embedded']['projects']) == 1




#!/usr/bin/env python
import os
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['easyapi/tests', '-s']
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(
    name='easy-rest',
    version='0.4.2',
    author='mturilin',
    author_email='mturilin@gmail.com',
    description='Real Python Enums for Django.',
    license='MIT',
    url='https://github.com/mturilin/easyapi',
    long_description='',
    packages=find_packages(),
    zip_safe=False,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    install_requires=[
        'django',
        'djangorestframework',
        'python-dateutil',
        'isodate',
        'six',
        'enum34',
        'django-enumfields',
    ],
    tests_require=[
        'pytest',
        'pytest-django',
        'factory-boy',
    ],
    cmdclass={'test': PyTest},
)



import json

import pytest
from rest_framework.status import HTTP_200_OK, HTTP_201_CREATED

from easyapi.tests.factories import CompanyFactory, ProjectFactory, ManagerFactory


__author__ = 'mikhailturilin'


@pytest.mark.django_db
def test_foreign_keys_end_with_id(staff_api_client):
    company = CompanyFactory()
    projects = [ProjectFactory(company=company) for i in range(3)]

    response = staff_api_client.get('/api/project/')
    assert response.status_code == HTTP_200_OK

    response_data = json.loads(response.content)

    for proj_dict in response_data:
        assert 'company_id' in proj_dict
        assert proj_dict['company_id'] == company.id


@pytest.mark.django_db
def test_create_with_foreign_keys_end_with_id(staff_api_client):
    company = CompanyFactory()
    manager = ManagerFactory()

    response = staff_api_client.post('/api/project/',
                                     {'name': 'aaaa', 'company_id': company.id, 'manager_id': manager.id,
                                      'start_date': '2014-05-19', 'scope': "Company"})
    assert response.status_code == HTTP_201_CREATED
    response_data = json.loads(response.content)

    proj_dict = response_data
    assert 'company_id' in proj_dict
    assert proj_dict['company_id'] == company.id





from easyapi.tests.test_project.models import Project, Company, Category, CompanyType, ProjectScope, Address, Manager
from datetime import timedelta, date
from django.utils import timezone

import factory
from factory import django as factory_django
from factory import fuzzy

__author__ = 'mikhailturilin'


class CategoryFactory(factory_django.DjangoModelFactory):
    class Meta:
        model = Category

    name = fuzzy.FuzzyText()


class AddressFactory(factory_django.DjangoModelFactory):
    class Meta:
        model = Address

    street = fuzzy.FuzzyText()

class CompanyFactory(factory_django.DjangoModelFactory):
    class Meta:
        model = Company

    category = factory.SubFactory(CategoryFactory)
    name = fuzzy.FuzzyText()
    country = fuzzy.FuzzyText(length=20)
    company_type = fuzzy.FuzzyChoice(CompanyType)
    address = None


SIX_MONTH_EARLIER = date.today() - timedelta(days=int(30.5*6))

class ManagerFactory(factory_django.DjangoModelFactory):
    class Meta:
        model = Manager

    name = fuzzy.FuzzyText()

class ProjectFactory(factory_django.DjangoModelFactory):
    class Meta:
        model = Project

    company = factory.SubFactory(CompanyFactory)
    name = fuzzy.FuzzyText()
    budget = fuzzy.FuzzyInteger(1000,9000)
    start_date = fuzzy.FuzzyDate(start_date=SIX_MONTH_EARLIER)
    scope = fuzzy.FuzzyChoice(ProjectScope)
    manager = factory.SubFactory(ManagerFactory)


from easyapi.router import EasyApiRouter, AutoAppRouter, AutoAppListRouter
from easyapi.tests.test_project.models import Company, Project
from easyapi.viewsets import InstanceViewSet

__author__ = 'mikhailturilin'

class CompanyViewSet(InstanceViewSet):
    model = Company

class ProjectViewSet(InstanceViewSet):
    model = Project

router = EasyApiRouter()
router.register('company', CompanyViewSet)
router.register('project', ProjectViewSet)



